//OpenGL introduction TP (in english parce que ca fait plus mieux)
//Your best friend : http://www.opengl.org/ :)

//Windows header
#ifdef WIN32
#include <windows.h>
#endif

// GLEW header for OpenGL extentions loading
#include "GL/glew.h"
// OpenGL header
#include <GL/gl.h>
// Glut for display window managment
#include <gl/glut.h>

#include "NemoGraphics/Vector.h"
#include "NemoGraphics/Matrix.h"

#include <stdio.h>
#include <string>

#include "Helpers.h"


using namespace NemoGraphics;

float xrot;
float yrot;
float xspeed;			// X Rotation Speed
float yspeed;			// Y Rotation Speed


bool lightMode;

Vector2i windowSize;

Vector3f viewPos(0.0, -2.0, -5.0);
Vector3f viewOrient;

Vector3f lightPos(2.0, 2.0, 2.0);

float rotAngle=0.0;


///Matrices///
Mat4f projectionMatrix; //Projection matrix
Mat4f modelWorldMatrix; //Transformation matrix from model to world
Mat4f worldViewMatrix;	//Transformation from world to view

///Shaders///
GLuint shaderProgramID;

//Create GLSL shader program (defined in ShadersManagemnt.cpp)
GLuint createShaderProgram(const char *fileNameVS, const char *fileNameFS, GLuint programID=0);

//Pass parameters data to vertex shader
void initLemming(){
		//Vertices positions
		//Get attribute ID
		GLuint attribID=glGetAttribLocation( shaderProgramID, "vertexPosition" );
		//Connect data array to attribute
		glVertexAttribPointer( attribID, 
			4, GL_FLOAT, GL_FALSE, 0, getModelVertices(0));
		//Attribute array will be used for rendering
		glEnableVertexAttribArray( attribID );

		//Vertices normals
		//Get attribute ID
		attribID=glGetAttribLocation( shaderProgramID, "vertexNormal" );
		//Connect data array to attribute
		glVertexAttribPointer( attribID, 
			3, GL_FLOAT, GL_FALSE, 0, getModelNormals(0));
		//Attribute array will be used for rendering
		glEnableVertexAttribArray( attribID );

		//Vertices colors
		//TODO: pass colors
}



//OpenGL initialisation function
void init(void) {
  // Initialise extensions loader.
  glewInit();

  // Create shader program from vertex and fragment shaders sources
  shaderProgramID=createShaderProgram("OpenGLTPbase_VS.glsl", "OpenGLTPbase_FS.glsl", 0);

  lightMode=false;

  initModel();

  ///Input primitives data definition///
	initLemming();
}


//Called when window size changed (window width, window height)
void reshape( int w, int h ) {
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	float ratio = float(w) / float(h);
	//Setup the projection matrix
	projectionMatrix.perspective(60.0f, ratio, 0.010f, 20.0f);
	//TODO: remplacer par une projection orthographique

	// Set the viewport to be the entire window
	// glViewport allow drawing to subparts of the display window
    glViewport(0, 0, w, h);
	//TODO: Fenetre de rendu centre et carre

	windowSize=Vector2i(w, h);
}


//Called each time the image must be redrawn
void display( void ) {
  glClearColor(0.1f, 0.1f, 0.2f, 0.5f);					// Set black clear color
  glClearDepth(1.0);									// Set initial depth buffer value to 1.0
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT ); // Effective clearing of the buffers

  ////HERE WE HAVE BEAUTIFUL BLACK MONOCHROM IN THE FRAME BUFFER////

  ///Drawing commands///

  //Depth buffer
  glDisable(GL_DEPTH_TEST);
  //glEnable(GL_DEPTH_TEST);
  //glDepthFunc(...
  //glDepthMask(...

  //Disable Blending
  glDisable(GL_BLEND);

  // Enable backface culling
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);

  //First enable shader program
  glUseProgram(shaderProgramID);

  //Build transformation matrix
  Mat4f modelViewMatrix = Mat4f::identity();
  modelViewMatrix *= Mat4f::rotation(Vector3f(1.0, 0.0, 0.0), viewOrient.x);
  modelViewMatrix *= Mat4f::rotation(Vector3f(0.0, 1.0, 0.0), viewOrient.y);
	modelViewMatrix *= Mat4f::translation(viewPos);



  //Pass matrices to shader
  glUniformMatrix4fv(glGetUniformLocation(shaderProgramID, "projectionMat"), 1, GL_FALSE, projectionMatrix.mat);
  glUniformMatrix4fv(glGetUniformLocation(shaderProgramID, "modelViewMat"), 1, GL_FALSE, modelViewMatrix.mat);
	//glUniformMatrix4fv(glGetUniformLocation(shaderProgramID, "modelViewMatIT"), 1, GL_FALSE, modelViewMatrix.inverse().transpose().mat);



	////HERE WE HAVE NOTHING BUT BLACK IN THE FRAME BUFFER////

		//Ask OpenGL for drawing primitives. Enabled attributes arrays will be used.
		//Indices stored in "indices" array are used as index in the attributes arrays.
		glDrawElements(GL_TRIANGLES, getModelNumIndices(0), GL_UNSIGNED_SHORT, getModelIndices(0));

		////HERE THE LEMMING HAVE BEEN DRAWN IN THE FRAME BUFFER////
		//The vertices stored in vertex array have been passing through all the graphics pipeline stages and finaly becomes pixels in the framebuffer.


		//Question 4//
		//Second passe: draw mesh outline
		//glLineWidth(4.0);
		//glCullFace(...
		//glPolygonMode(GL_BACK, GL_LINE);
		//glDrawElements(...
		//glPolygonMode(GL_BACK, GL_FILL);
		//glCullFace(...


	//Disable shader program
	glUseProgram(0);

  xrot+=xspeed;
  yrot+=yspeed;
	rotAngle+=5.0;

  //Swap front and back buffer -> display drawn elements
  glutSwapBuffers();
}


Vector3f movement;
float speed=0.05f;
void keyboard ( unsigned char key, int x, int y )  // Create Keyboard Function
{
	movement=Vector3f(0.0);

  switch ( key ) {
  case 27:        // When Escape Is Pressed...
    exit ( 0 );   // Exit The Program
    break;
	case 'l':
		lightMode=!lightMode;
    break;
	case 'z':     // When Up Arrow Is Pressed...
		movement.z+=speed;
		break;
	case 's':               // When Down Arrow Is Pressed...
		movement.z-=speed;
		break;
	case 'q':
		movement.x+=speed;
		break;
	case 'd':
		movement.x-=speed;
		break;
	case 'r':
		 createShaderProgram("OpenGLTPbase_VS.glsl", "OpenGLTPbase_FS.glsl", shaderProgramID);
		 break;
   default:
      break;
  }

	movement.rotate(Vector3f(1.0, 0.0, 0.0), -viewOrient.x*PI/180.0);
	movement.rotate(Vector3f(0.0, 1.0, 0.0), -viewOrient.y*PI/180.0);

	viewPos=viewPos+movement;
}

void arrow_keys ( int a_keys, int x, int y )  // Create Special Function (required for arrow keys)
{
  switch ( a_keys ) {
	case GLUT_KEY_F1:
		glutFullScreen ();
		break;
	case GLUT_KEY_F2:
		glutReshapeWindow ( 500, 500 );
		break;
	default:
		break;
  }
}

bool mclicked;
Vector2i moldpos;

//Called when a mouse button is pressed or released
void mouse_click(int button, int state, int x, int y){
	mclicked=state;
	moldpos=Vector2i(x, y);
}

//Called when a mouse move and a button is pressed
void mouse_motion(int x, int y){
	Vector2i mmotion=Vector2i(x, y)-moldpos;
	Vector2f mmotionf=Vector2f(mmotion)/Vector2f(windowSize);

	if(!lightMode){
		viewOrient.x=viewOrient.x+mmotionf.y*100.0;
		viewOrient.y=viewOrient.y+mmotionf.x*100.0;
	}else{
		lightPos.rotate(Vector3f(1.0, 0.0, 0.0), mmotionf.y*5.0);
		lightPos.rotate(Vector3f(0.0, 1.0, 0.0), mmotionf.x*5.0);
	}

	moldpos=Vector2i(x, y);
}


void main ( int argc, char** argv )   // Create Main Function For Bringing It All Together
{
  glutInit            ( &argc, argv ); // Erm Just Write It =)
  glutInitDisplayMode ( GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA | GLUT_STENCIL); // Display Mode
  glutInitWindowPosition (0,0);
  glutInitWindowSize  ( 500, 500 ); // If glutFullScreen wasn't called this is the window size
  glutCreateWindow    ( "M2R OpenGL TP" ); // Window Title (argv[0] for current directory as title)
  init ();

  glutDisplayFunc     ( display );  // Matching Earlier Functions To Their Counterparts
  glutReshapeFunc     ( reshape );
  glutKeyboardFunc    ( keyboard );
  glutSpecialFunc     ( arrow_keys );
	glutMotionFunc	    ( mouse_motion );
	glutMouseFunc	    ( mouse_click );
  glutIdleFunc			 ( display );
  glutMainLoop        ( );          // Initialize The Main Loop
}
